--[[
    Polling for interrupts
]]

local addon_name, ns = ...

local Printer = LibStub:GetLibrary("OakPrint-1.0")(addon_name)

local specTable = {}
local classTable = {}

local tContains, tinsert, CheckInteractDistance = tContains, tinsert, CheckInteractDistance

--------------------------------------------------------------------------------
-- Dataobject
--------------------------------------------------------------------------------

local ldb = LibStub:GetLibrary("LibDataBroker-1.1")
local dataobj = ldb:NewDataObject("OakInterrupts",
        {
            type = "OakInterrupt",
            polling = false
        })

--------------------------------------------------------------------------------
-- Helper functions/tables
--------------------------------------------------------------------------------

local mt = {
    __index = function(table, key)
        if key == "cooldown" then return 15
        else return nil
        end
    end
}

local spellTable = {
    [2139]   = { cooldown = 24 },
    [96231]  = { cooldown = 15 },
    [6552]   = { cooldown = 15 },
    [78675]  = { cooldown = 60 },
    [106839] = { cooldown = 15 },
    [47528]  = { cooldown = 15 },
    [147362] = { cooldown = 24 },
    [187707] = { cooldown = 15 },
    [15487]  = { cooldown = 45 },
    [1766]   = { cooldown = 15 },
    [57994]  = { cooldown = 12 },
    [116705] = { cooldown = 15 },
    [183752] = { cooldown = 15 }
}

local spellOfSpec = {
    [62]  = 2139,
    [63]  = 2139,
    [64]  = 2139,
    [66]  = 96231,
    [70]  = 96231,
    [71]  = 6552,
    [72]  = 6552,
    [73]  = 6552,
    [102] = 78675,
    [103] = 106839,
    [104] = 106839,
    [250] = 47528,
    [251] = 47528,
    [252] = 47528,
    [253] = 147362,
    [254] = 147362,
    [255] = 187707,
    [258] = 15487,
    [259] = 1766,
    [260] = 1766,
    [261] = 1766,
    [262] = 57994,
    [263] = 57994,
    [264] = 57994,
    [268] = 116705,
    [269] = 116705,
    [577] = 183752,
    [581] = 183752
}

local spellOfClass = {
    ["MAGE"] = 2139,
    ["WARRIOR"] = 6552,
    ["DEATHKNIGHT"] = 46528,
    ["ROGUE"] = 1766,
    ["SHAMAN"] = 57994,
    ["DEMONHUNTER"] = 183752
}

local rangedSpec = {
    [62] = true,
    [63] = true,
    [64] = true,
    [102] = true,
    [253] = true,
    [254] = true,
    [258] = true,
    [262] = true,
    [264] = true
}

local filterGroups = {
    ["RANGED"] = rangedSpec
}

local specClass = {
    [62] = "MAGE",
    [63] = "MAGE",
    [64] = "MAGE",
    [65] = "PALADIN",
    [66] = "PALADIN",
    [70] = "PALADIN",
    [71] = "WARRIOR",
    [72] = "WARRIOR",
    [73] = "WARRIOR",
    [102] = "DRUID",
    [103] = "DRUID",
    [104] = "DRUID",
    [105] = "DRUID",
    [250] = "DEATHKNIGHT",
    [251] = "DEATHKNIGHT",
    [252] = "DEATHKNIGHT",
    [253] = "HUNTER",
    [254] = "HUNTER",
    [255] = "HUNTER",
    [256] = "PRIEST",
    [257] = "PRIEST",
    [258] = "PRIEST",
    [259] = "ROGUE",
    [260] = "ROGUE",
    [261] = "ROGUE",
    [262] = "SHAMAN",
    [263] = "SHAMAN",
    [264] = "SHAMAN",
    [265] = "WARLOCK",
    [266] = "WARLOCK",
    [267] = "WARLOCK",
    [268] = "MONK",
    [269] = "MONK",
    [270] = "MONK",
    [577] = "DEMONHUNTER",
    [581] = "DEMONHUNTER"
}

--[[
local specs = {
    -- Mage --
    [62] = "Arcane",
    [63] = "Fire",
    [64] = "Frost",
    -- Paladin --
    [65] = "Holy",
    [66] = "Protection",
    [70] = "Retribution",
    -- Warrior --
    [71] = "Arms",
    [72] = "Fury",
    [73] = "Protection",
    -- Druid --
    [102] = "Balance",
    [103] = "Feral",
    [104] = "Guardian",
    [105] = "Restoration",
    -- Death Knight --
    [250] = "Blood",
    [251] = "Frost",
    [252] = "Unholy",
    -- Hunter --
    [253] = "Beast Mastery",
    [254] = "Marksmanship",
    [255] = "Survival",
    -- Priest --
    [256] = "Discipline",
    [257] = "Holy",
    [258] = "Shadow",
    -- Rogue --
    [259] = "Assassination",
    [260] = "Outlaw",
    [261] = "Subtlety",
    -- Shaman --
    [262] = "Elemental",
    [263] = "Enhancement",
    [264] = "Restoration",
    -- Warlock --
    [265] = "Affliction",
    [266] = "Demonology",
    [267] = "Destruction",
    -- Monk --
    [268] = "Brewmaster",
    [269] = "Windwalker",
    [270] = "Mistweaver",
    -- Demon Hunter --
    [577] = "Havoc",
    [581] = "Vengence"
}
]]

local irrelevantSpec = {
    [65] = true,
    [105] = true,
    [256] = true,
    [257] = true,
    [265] = true,
    [266] = true,
    [267] = true,
    [270] = true
}

local shortcutFilter = {}
local specFilter = setmetatable({}, { __index = irrelevantSpec })
-- default filter
do
    for k, _ in pairs(rangedSpec) do
        specFilter[k] = true
        shortcutFilter[specClass[k]] = true
    end
end

local function tremovebyval(t, val)
    for k, v in ipairs(t) do
        if v == val then
            tremove(t, k)
            break
        end
    end
end

--------------------------------------------------------------------------------
-- Metatable magic
--------------------------------------------------------------------------------

local function IsPlayerGUID(str)
    return str and string.match(str, 'Player-.*')
end


local nameTable = {}

setmetatable(nameTable, {
    __index = function(t, k)
        if k ~= nil then
            name = IsPlayerGUID(k) and select(6, GetPlayerInfoByGUID(k)) or nil
            t[k] = name
            return name
        end
    end})

local _interruptTable = {}
local interruptTable = {}

setmetatable(interruptTable, {
    __index = _interruptTable,
    __newindex = function(t, k, v)
        assert(not (k == nil), "interruptTable: tried to set nil key")
        local name = nameTable[k]
        if v == nil then -- unregister
            _interruptTable[k] = nil
            dataobj[name] = nil
            return
        end

        local class = classTable[k]
        local spellID = spellOfClass[class] or spellOfSpec[specTable[k]]
        local spell = spellTable[spellID]

        if spell then
            local cd = spell.cooldown
            local time = v > 0 and v + cd or _interruptTable[k] or 0
            Printer.debug(2,
                       "Setting interrupt for %s with expiration time %d",
                       name, time)
            _interruptTable[k] = time
            dataobj[name] = time -- set exported data
        end
    end,
    __call = function()
        return pairs(_interruptTable)
    end
})

--------------------------------------------------------------------------------
-- Inspection
--------------------------------------------------------------------------------

local notifyRegister = {}
local oorRegister = {}
local inspecting
local inspect_timeout

local function RegisterInspection(name)
    Printer.debug(2, "Registering %s for inspection", name)
    if not tContains(notifyRegister, name) then
        tinsert(notifyRegister, name)
    end
    if not inspecting then
        Printer.debug(3, "Next inspection is for %s (calling NotifyInspect)", name)
        NotifyInspect(name)
        inspect_timeout = 4
        inspecting = UnitGUID(name)
    end
end

local next_oorpass = 8
local function OoRInspection(self, elapsed)
    if elapsed > next_oorpass then
        local switching = {}
        for _, v in ipairs(oorRegister) do
            if CheckInteractDistance(v, 1) then
                tinsert(switching, v)
            end
        end
        for _, v in ipairs(switching) do
            tremovebyval(oorRegister, v)
            RegisterInspection(v)
        end
        next_oorpass = 8
    else
        next_oorpass = next_oorpass - elapsed
    end
end

local function InspectNext()
    for k, name in ipairs(notifyRegister) do
        if CheckInteractDistance(name, 1) then
            RegisterInspection(name)
            break
        else
            Printer.debug(2, "%s is too far away to inspect", name)
            if not tContains(oorRegister, name) then
                tinsert(oorRegister, 1, name)
            end
        end
    end
    for k, name in ipairs(oorRegister) do
        tremovebyval(notifyRegister, name)
    end
end

local function InspectGUID(guid)
    local name = nameTable[guid]
    assert(name, "No name associated with GUID " .. guid)
    Printer.debug(2, "Name: %s, GUID: %s", name, guid)

    local specID = GetInspectSpecialization(name)
    specTable[guid] = specID
    if specFilter[specID] then
        Printer.debug(2, "%s has irrelevant spec", name)
        interruptTable[guid] = nil --"UNREGISTER"
    else
        if not interruptTable[guid] then
            interruptTable[guid] = 0
        end
        oorRegister[name] = nil
    end
    tremovebyval(notifyRegister, name)
    ClearInspectPlayer()
    inspecting = nil
end

--------------------------------------------------------------------------------
-- Polling
--------------------------------------------------------------------------------

local function PollByUnitID(unitID, force)
    if not UnitIsPlayer(unitID) then return end
    local name = UnitName(unitID)
    local guid = UnitGUID(unitID)
    if not guid or not name or name == "Unknown" then return end
    if force or not interruptTable[guid] then
        Printer.debug(2, "Processing %s:%s", name, guid)
        local class = select(2, UnitClass(unitID))
        if class == "WARLOCK" then
            Printer.debug(2, "%s is a warlock, we don't like warlocks")
            return
        end
        classTable[guid] = class
        if spellOfClass[class] and not shortcutFilter[class] then
            -- all specs of the class have the same interrupt
            interruptTable[guid] = 0
        else
            RegisterInspection(name)
        end
    end
end

local function TestPoll()
    Printer.debug(1, "Running test poll")
    PollByUnitID('player', true)
end

local function PurgeUnit(guid)
    local name = nameTable[guid]
    Printer.debug(3, "Purging %s:%s", name, guid)
    interruptTable[guid] = nil
    tremovebyval(notifyRegister, name)
    tremovebyval(oorRegister, name)
end

local function PollGroup(force)
    Printer.debug(1, "Clearing stale data")
    for guid, _ in interruptTable() do
        local name = nameTable[guid]
        if force or not UnitInRaid(name) and not UnitInParty(name) then
            PurgeUnit(guid)
        end
    end
    for _, name in pairs(oorRegister) do
        if force or not UnitInRaid(name) and not UnitInParty(name) then
            tremovebyval(oorRegister, name)
        end
    end
    if IsInGroup() then
        Printer.debug(1, "Polling group")
        local num = GetNumGroupMembers()
        local prefix

        if IsInRaid() then
            prefix = 'raid'
        else
            prefix = 'party'
            PollByUnitID('player')
            num = num - 1
        end

        for i=1,num do
            local unit = prefix .. i
            PollByUnitID(unit)
        end
    end
    inspect_timeout = 1
end

local function filter(mode, ...)
    if mode == 'reset' then
        specFilter = setmetatable({}, { __index = irrelevantSpec })
    end

    local numFlags = select('#', ...)
    for i=1, numFlags do
        local flag = select(i, ...)
        if flag then
            for k, v in pairs(filterGroups[string.upper(flag)]) do
                specFilter[k] = mode and v or not v
            end
        end
    end
    shortcutFilter = {}
    for k, v in pairs(specFilter) do
        shortcutFilter[specClass[k]] = shortcutFilter[specClass[k]] or specFilter[k]
    end

    PollGroup(true)
end

--------------------------------------------------------------------------------
-- Controller
--------------------------------------------------------------------------------

local controller = CreateFrame("Frame")

controller:RegisterEvent("INSPECT_READY")
controller:RegisterEvent("UNIT_SPELLCAST_SUCCEEDED")
controller:RegisterEvent("GROUP_ROSTER_UPDATE")
controller:RegisterEvent("PLAYER_SPECIALIZATION_CHANGED")

function controller.OnUpdate(self, elapsed)
    if inspecting then
        if inspect_timeout and inspect_timeout < elapsed then
            Printer.debug(3, "Inspection timed out, restarting...")
            ClearInspectPlayer()
            inspecting = nil
            InspectNext()
        else
            inspect_timeout = inspect_timeout - elapsed
        end
    end
    if #oorRegister > 0 then OoRInspection(self, elapsed) end
    if #oorRegister == 0 and #notifyRegister == 0 then
        Printer.debug(2, 'Nothing left to inspect, unregistering update script')
        controller:SetScript('OnUpdate', nil)
        dataobj.polling = false
    end
end


local function handleEvents(self, event, ...)
        Printer.debug(5, "Event caught: %s\n    args: %s", event, strjoin(", ", ...))
        if event == "UNIT_SPELLCAST_SUCCEEDED" then
            local unitID, spell, _, lineID, spellID = ...
            local guid = UnitGUID(unitID)
            if type(interruptTable[guid]) == "number" and spellTable[spellID] then
                Printer.debug(5, "Tracked spell (%s) cast by %s", spell, nameTable[guid])
                interruptTable[guid] = GetTime()
            end
        elseif event == "INSPECT_READY" then
            local guid = ...
            if inspecting == guid then
                InspectGUID(guid)
                InspectNext()
            else
                Printer.debug(3, "%s is not the current inspection target", UnitName(guid) or guid)
            end
        elseif event == "GROUP_ROSTER_UPDATE" then
            Printer.debug(2, "Roster Update")
            if IsInGroup() and not controller:GetScript('OnUpdate') then
                Printer.debug(3, "Registering update script")
                controller:SetScript('OnUpdate', controller.OnUpdate)
                dataobj.polling = true
            elseif not IsInGroup() then
                Printer.debug(3, "Left group, unregistering update script")
                controller:SetScript('OnUpdate', nil)
            end
            PollGroup()
        elseif event == "PLAYER_SPECIALIZATION_CHANGED" then
            local unitID = ...
            if not unitID then return end
            Printer.debug(2, "%s changed spec", unitID)
            if UnitInRaid(unitID) or UnitInParty(unitID) then
                PollByUnitID(unitID, true)
            end
        end
    end

controller:SetScript("OnEvent", handleEvents)

--------------------------------------------------------------------------------
-- Slash commands
--------------------------------------------------------------------------------

local function Dump(heading, t)
    if t == dataobj then
        _, t = ldb:pairs(t)
    end
    Printer.print("--- " .. heading .. " ---")
    for k, v in next, t do
       Printer.print("%s: %s", k, tostring(v))
    end
end

local misc_table = {}
local function UpdateMiscTable()
    misc_table["Update script"] = controller:GetScript('OnUpdate') and 'set'
                                        or 'not set'
    misc_table["Inspection target"] = nameTable[inspecting] or 'none'
end

local commands = LibStub:GetLibrary("OakSlash-1.0")(addon_name)
local t = {
    poll = PollGroup,
    filter = { ranged = function(...) filter(true, 'ranged') end,
               none = function() filter('reset') end},
    test = TestPoll,
    dump = { all = function()
                       UpdateMiscTable()
                       Dump("Miscellaneous", misc_table)
                       Dump("Exported Data", dataobj)
                       Dump("Notify Register", notifyRegister)
                       Dump("Out of Range", oorRegister)
                   end,
             data = function() Dump("Exported Data", dataobj) end,
             inspect = function()
                           Dump("Notify Register", notifyRegister)
                           Dump("Out of Range", oorRegister)
                       end,
             misc = function()
                        UpdateMiscTable()
                        Dump("Miscellaneous", misc_table)
                    end},
    debug = Printer.SetDebug
}
commands:NewCommand("oib", t)
